Unit MKComDir;

{Based heavily on Wayne Conrad's IbmCom Version 3.2}
{Much thanks to him for making it available}

{
     MKComDir - Copyright 1993 by Mark May - MK Software
     You are free to use this code in your programs, however
     it may not be included in Source/TPU function libraries
     without my permission.

     Mythical Kingom Tech BBS (513)237-7737 HST/v32
     FidoNet: 1:110/290
     Rime: ->MYTHKING
     You may also reach me at maym@dmapub.dma.org
}


{$R-}
{$I MKB.Def}


Interface

Uses MKComAbs,
  {$IFDEF WINDOWS}
  WinDos;
  {$ELSE}
  Dos;
  {$ENDIF}

Type DirComObj = Object(AbsComObj)
  Constructor Init;
  Destructor Done; Virtual;
  Function OpenPort(PortNo: Byte; Baud: LongInt; Flow: Byte; NumBits: Byte;
    StopBits: Byte; Parity: Byte): Boolean; Virtual;
  Procedure ClosePort; Virtual;
  Procedure SetBPS(Rate: LongInt); Virtual;
  Function  GetBPS: LongInt; Virtual;
  Procedure LowerDTR; Virtual;
  Procedure RaiseDTR; Virtual;
  Function  GetInfo: String; Virtual;
  Function  OutReady: Boolean; Virtual;
  Function  OutEmpty: Boolean; Virtual;
  Function  InEmpty: Boolean; Virtual;
  Function  GetChar: Char; Virtual;
  Function  PeekChar: Char; Virtual;
  Procedure SendChar(Ch: Char); Virtual;
  Function  SendCharNW(Ch: Char): Boolean; Virtual;
  Procedure SetFlowCtl(FC: Byte); Virtual;
  Procedure StartBreak; Virtual;
  Procedure StopBreak; Virtual;
  Function  SendBlock(Var Dt; DSize: Word): Word; Virtual;
  Function  GetBlock(Var Dt; DSize: Word): Word; Virtual;
  Function  GetCarrier: Boolean; Virtual;
  Procedure FlushIn; Virtual;
  Procedure FlushOut; Virtual;
  End;


Type DirComPtr = ^DirComObj;


Implementation

{This unit is the communications port interrupt driver for the IBM-PC.
It handles handles all low-level i/o through the serial port.  It is
installed by calling com_install.  It deinstalls itself automatically
when the program exits, or you can deinstall it by calling com_deinstall.

Donated to the public domain by Wayne E. Conrad, January, 1989.
If you have any problems or suggestions, please contact me at my BBS:

    Pascalaholics Anonymous
    (602) 484-9356
    2400 bps
    The home of WBBS
    Lots of source code

Version 3.0, January, 1989:
  Initial Release to the public domain.

Version 3.1, soon after version 3.0:
  Corrected my BBS number.  Oops.

Version 3.2, January, 1990:
  Added CTS (request-to-send) flow checking.  The driver will not
  transmit unless CTS is active.  This is for high-speed modems and
  modems with MNP, such as the US Robotics HST.  Made some extremely
  minor changes to comments.
}


{Note some modifications and OOP extensions 1992, 1993 by Mark May}
{Mythical Kingdom Tech Bbs 513-237-7737 1:110/290}


Type
  com_parity = (com_none, com_even, com_odd, com_zero, com_one);


{Summary of IBM-PC Asynchronous Adapter Registers.  From:
  Compute!'s Mapping the IBM PC and PCjr, by Russ Davis
  (Greensboro, North Carolina, 1985: COMPUTE! Publications, Inc.),
  pp. 290-292.

Addresses given are for COM1 and COM2, respectively.  The names given
in parentheses are the names used in this module.


3F8/2F8 (uart_data) Write: transmit buffer.  Read: receive buffer, or
baud rate divisor LSB if port 3FB, bit 7 = 1.

3F9/2F9 (uart_ier) Write: Interrupt enable register or baud rate divisor
MSB if port 3FB, bit 7 = 1.
PCjr baud rate divisor is different from other models;
clock input is 1.7895 megahertz rather than 1.8432 megahertz.
Interrupt enable register:
    bits 7-4  forced to 0
    bit 3     1=enable change-in-modem-status interrupt
    bit 2     1=enable line-status interrupt
    bit 1     1=enable transmit-register-empty interrupt
    bit 0     1=data-available interrupt

3FA/2FA (uart_iir) Interrupt identification register (prioritized)
     bits 7-3  forced to 0
     bits 2-1  00=change-in-modem-status (lowest)
     bits 2-1  01=transmit-register-empty (low)
     bits 2-1  10=data-available (high)
     bits 2-1  11=line status (highest)
     bit 0     1=no interrupt pending
     bit 0     0=interrupt pending

3FB/2FB (uart_lcr) Line control register
     bit 7  0=normal, 1=address baud rate divisor registers
     bit 6  0=break disabled, 1=enabled
     bit 5  0=don't force parity
            1=if bit 4-3=01 parity always 1
              if bit 4-3=11 parity always 0
              if bit 3=0 no parity
     bit 4  0=odd parity,1=even
     bit 3  0=no parity,1=parity
     bit 2  0=1 stop bit
            1=1.5 stop bits if 5 bits/character or
              2 stop bits if 6-8 bits/character
     bits 1-0  00=5 bits/character
               01=6 bits/character
               10=7 bits/character
               11=8 bits/character

     bits 5..3: 000 No parity
                001 Odd parity
                010 No parity
                011 Even parity
                100 No parity
                101 Parity always 1
                110 No parity
                111 Parity always 0


3FC/2FC (uart_mcr) Modem control register
     bits 7-5  forced to zero
     bit 4     0=normal, 1=loop back test
     bits 3-2  all PCs except PCjr
     bit 3     1=interrupts to system bus, user-designated output: OUT2
     bit 2     user-designated output, OUT1
     bit 1     1=activate rts
     bit 0     1=activate dtr

3FD/2FD (uart_lsr) Line status register
     bit 7  forced to 0
     bit 6  1=transmit shift register is empty
     bit 5  1=transmit hold register is empty
     bit 4  1=break received
     bit 3  1=framing error received
     bit 2  1=parity error received
     bit 1  1=overrun error received
     bit 0  1=data received

3FE/2FE (uart_msr) Modem status register
     bit 7  1=receive line signal detect
     bit 6  1=ring indicator (all PCs except PCjr)
     bit 5  1=dsr
     bit 4  1=cts
     bit 3  1=receive line signal detect has changed state
     bit 2  1=ring indicator has changed state (all PCs except PCjr)
     bit 1  1=dsr has changed state
     bit 0  1=cts has changed state

3FF/2FF (uart_spr) Scratch pad register.}


{Maximum port number (minimum is 1) }

Const
  max_port = 4;


{Base i/o address for each COM port}

Const
  uart_base: Array [1..max_port] OF Integer = ($3F8, $2F8, $3E8, $2E8);


{Interrupt numbers for each COM port}

Const
  intnums: Array [1..max_port] OF Byte = ($0C, $0B, $0C, $0B);


{i8259 interrupt levels for each port}

Const
  i8259levels: Array [1..Max_Port] OF Byte = (4, 3, 4, 3);


{This Variable is TRUE if the interrupt driver has been installed, or
FALSE if it hasn't.  It's used to prevent installing twice or
deinstalling when not installed.}

Const
  Com_Installed: Boolean = False;


{UART i/o addresses.  Values depend upon which COMM port is selected.}

Var
  uart_data: Word;             {Data register}
  uart_ier : Word;             {Interrupt enable register}
  uart_iir : Word;             {Interrupt identification register}
  uart_lcr : Word;             {Line control register}
  uart_mcr : Word;             {Modem control register}
  uart_lsr : Word;             {Line status register}
  uart_msr : Word;             {Modem status register}
  uart_spr : Word;             {Scratch pad register}


{Original contents of IER and MCR registers.  Used to restore UART to
whatever state it was in before this driver was loaded.}

Var
  Old_Ier: Byte;
  Old_Mcr: Byte;


{Original contents of interrupt vector.  Used to restore the vector when
the interrupt driver is deinstalled.}

Var
  Old_Vector: Pointer;


{Original contents of interrupt controller mask.  Used to restore the bit
pertaining to the comm controller we're using.}

Var
  old_i8259_mask: Byte;


{Bit mask for i8259 interrupt controller}

Var
  i8259bit: Byte;


{Interrupt vector number}

Var
  intnum: Byte;


{Receive queue.  Received characters are held here until retrieved by
com_rx.}

Const
  rx_queue_size = 512;   {Change to suit}
  tx_queue_size = 512;    {Change to suit}

Type Rx_QueueType = Array[1..Rx_Queue_Size] of Byte;
Type Tx_QueueType = Array[1..Tx_Queue_Size] of Byte;

Var
  rx_queue: ^Rx_QueueType;
  rx_in   : Word;        {Index of where to store next character}
  rx_out  : Word;        {Index of where to retrieve next character}
  rx_chars: Word;        {Number of chars in queue}
  tx_queue: ^Tx_QueueType;
  tx_in   : Word;        {Index of where to store next character}
  tx_out  : Word;        {Index of where to retrieve next character}
  tx_chars: Word;        {Number of chars in queue}


{This Variable is used to save the next link in the "exit Procedure" chain.}
Var
  exit_save: Pointer;


Procedure disable_interrupts;
Inline
  (
  $FA    {CLI}
  );


Procedure enable_interrupts;
Inline
  (
  $FB    {STI}
  );




{Interrupt driver.  The UART is programmed to cause an interrupt whenever
a character has been received or when the UART is ready to transmit another
character.}


{$S-}
Procedure com_interrupt_driver; Interrupt;

Var
  ch   : Char;
  msr  : Byte;
  iir  : Byte;
  dummy: Byte;

Begin

  {While bit 0 of the interrupt identification register is 0, there is an
  interrupt to process}

  iir := Port [uart_iir];

  While Not Odd (iir) Do
    Begin
    Case ((iir Shr 1) and 3) Of

      {iir = 100b: Received data available.  Get the character, and if
      the buffer isn't full, then save it.  If the buffer is full,
      then ignore it.}

      2: Begin
         ch := Char (Port [uart_data] );
         If (rx_chars <= rx_queue_size) Then
           Begin
           rx_queue^ [rx_in] := Ord (ch);
           Inc (rx_in);
           IF rx_in > rx_queue_size THEN
             rx_in := 1;
           rx_chars := Succ (rx_chars);
           End;
         End;

      {iir = 010b: Transmit register empty.  If the transmit buffer is
      empty or CTS is not enabled, then disable the transmitter to
      prevent any more transmit interrupts.  Otherwise, send the
      character.

      The test of the line-status-register is to see if the transmit
      holding register is truly empty.  Some UARTS seem to cause transmit
      interrupts when the holding register isn't empty, causing
      transmitted characters to be lost.}

      1: Begin
         IF (tx_chars <= 0) OR NOT Odd (Port [uart_msr] SHR 4) Then
           Port [uart_ier] := Port [uart_ier] AND NOT 2
         Else
           IF Odd (Port [uart_lsr] SHR 5) Then
             Begin
             Port [uart_data] := tx_queue^ [tx_out];
             Inc (tx_out);
             IF tx_out > tx_queue_size Then
               tx_out := 1;
             Dec (tx_chars);
             End;
         End;

      {iir = 001b: Change in modem status.  If the CTS signal has changed
      to active, then enable the transmit interrupt.  If it has changed
      to inactive, then disable the transmit interrupt.}

      0: Begin
         msr := Port [uart_msr];
         IF Odd (msr) THEN
           IF Odd (msr SHR 4) THEN
             Port [uart_ier] := Port [uart_ier] OR 2
           ELSE
             Port [uart_ier] := Port [uart_ier] AND NOT 2
         End;

      {iir = 111b: Change in line status.  We don't expect this interrupt,
      but if one ever occurs we need to read the line status to reset it
      and prevent an endless loop.}

      3: Dummy := Port [uart_lsr];

      End;

    iir := Port [uart_iir];
    End;

  {Tell the interrupt controller that we're done with this interrupt}

  Port [$20] := $20;

  End;
{$S+}


Procedure DirComObj.FlushIn;
  Begin
  disable_interrupts;
  rx_chars := 0;
  rx_in    := 1;
  rx_out   := 1;
  enable_interrupts;
  End;


Procedure DirComObj.FlushOut;
  Begin
  disable_interrupts;
  tx_chars := 0;
  tx_in    := 1;
  tx_out   := 1;
  enable_interrupts;
  End;


Function DirComObj.GetCarrier: Boolean;
  Begin
  GetCarrier := com_installed AND Odd (Port [uart_msr] SHR 7);
  End;


Function DirComObj.GetChar: Char;
  Begin
  If Not com_installed Or (rx_chars = 0) Then
    GetChar := #0
  Else
    Begin
    disable_interrupts;
    GetChar := Chr (rx_queue^ [rx_out] );
    Inc (rx_out);
    If rx_out > rx_queue_size Then
      rx_out := 1;
    Dec (rx_chars);
    enable_interrupts;
    End;
  End;


Function DirComObj.PeekChar: Char;
  Begin
  If Not com_installed Or (rx_chars = 0) Then
    PeekChar := #0
  Else
    Begin
    disable_interrupts;
    PeekChar := Chr (rx_queue^ [rx_out] );
    enable_interrupts;
    End;
  End;


Function DirComObj.OutReady: Boolean;
  Begin
  OutReady := (tx_chars < tx_queue_size) OR NOT com_installed;
  End;


Function DirComObj.OutEmpty: Boolean;
  Begin
  OutEmpty := (tx_chars = 0) OR NOT com_installed;
  End;


Function DirComObj.InEmpty: Boolean;
  Begin
  InEmpty := (rx_chars = 0) Or Not com_installed;
  End;


Procedure DirComObj.SendChar(Ch: Char);
  Begin
  If com_installed Then
    Begin
    Repeat Until OutReady;
    disable_interrupts;
    tx_queue^ [tx_in] := Ord (ch);
    If tx_in < tx_queue_size Then
      Inc (tx_in)
    Else
      tx_in := 1;
    Inc (tx_chars);
    If Odd (Port [uart_msr] Shr 4) Then
      Port [uart_ier] := Port [uart_ier] Or 2;
    enable_interrupts;
    End;
  End;


Procedure DirComObj.LowerDTR;
  Begin
  If com_installed Then
    Begin
    disable_interrupts;
    Port [uart_mcr] := Port [uart_mcr] AND NOT 1;
    enable_interrupts;
    End;
  End;


Procedure DirComObj.RaiseDTR;
  Begin
  If com_installed Then
    Begin
    disable_interrupts;
    Port [uart_mcr] := Port [uart_mcr] OR 1;
    enable_interrupts;
    End;
  End;


Procedure DirComObj.SetBPS(Rate: LongInt);
  Var
    divisor: Word;
  Begin
  coBaud := Rate;
  If com_installed Then
    Begin
    If Rate < 2 Then
      Rate := 2;
    divisor := 115200 Div Rate;
    disable_interrupts;
    Port  [uart_lcr]  := Port [uart_lcr] OR $80;
    Portw [uart_data] := divisor;
    Port  [uart_lcr]  := Port [uart_lcr] AND NOT $80;
    enable_interrupts;
    End;
  End;


{Set the parity and stop bits as follows:

  com_none    8 data bits, no parity
  com_even    7 data bits, even parity
  com_odd     7 data bits, odd parity
  com_zero    7 data bits, parity always zero
  com_one     7 data bits, parity always one}


Procedure com_set_parity (parity: com_parity; stop_bits: Byte);
  Var
    lcr: Byte;

  Begin
  CASE parity OF
    com_none: lcr := $00 OR $03;
    com_even: lcr := $18 OR $02;
    com_odd : lcr := $08 OR $02;
    com_zero: lcr := $38 OR $02;
    com_one : lcr := $28 OR $02;
    End;
  IF stop_bits = 2 THEN
    lcr := lcr OR $04;
  disable_interrupts;
  Port [uart_lcr] := Port [uart_lcr] AND $40 OR lcr;
  enable_interrupts;
  End;


{Install the communications driver.  Portnum should be 1..max_port.
Error codes returned are:

  0 - No error
  1 - Invalid port number
  2 - UART for that port is not present
  3 - Already installed, new installation ignored}

Procedure com_install (portnum  : Word; Var error: Word);
  Var
    ier: Byte;

  Begin
  If com_installed Then
    Error := 3
  Else
    If (portnum < 1) OR (portnum > max_port) Then
      error := 1
    Else
      Begin
      {Set i/o addresses and other hardware specifics for selected port}
      uart_data := uart_base [portnum];
      uart_ier  := uart_data + 1;
      uart_iir  := uart_data + 2;
      uart_lcr  := uart_data + 3;
      uart_mcr  := uart_data + 4;
      uart_lsr  := uart_data + 5;
      uart_msr  := uart_data + 6;
      uart_spr  := uart_data + 7;
      intnum    := intnums [portnum];
      i8259bit  := 1 SHL i8259levels [portnum];
      {Return error if hardware not installed}
      old_ier := Port [uart_ier];
      Port [uart_ier] := 0;
      If Port [uart_ier] <> 0 Then
        Error := 2
      Else
        Begin
        Error := 0;
        {Save original interrupt controller mask, then disable the
        interrupt controller for this interrupt.}
        disable_interrupts;
        old_i8259_mask := Port [$21];
        Port [$21] := old_i8259_mask Or i8259bit;
        enable_interrupts;
        {Clear the transmit and receive queues}
        disable_interrupts;
        rx_chars := 0;
        rx_in    := 1;
        rx_out   := 1;
        tx_chars := 0;
        tx_in    := 1;
        tx_out   := 1;
        enable_interrupts;
        {Save current interrupt vector, then set the interrupt vector to
        the address of our interrupt driver.}
        GetIntVec (intnum, old_vector);
        SetIntVec (intnum, @com_interrupt_driver);
        com_installed := True;
        New(RX_Queue);
        New(TX_Queue);
        {Set parity to none, turn off BREAK signal, and make sure
        we're not addressing the baud rate registers.}
        Port [uart_lcr] := 3;
        {Save original contents of modem control register, then enable
        interrupts to system bus and activate RTS.  Leave DTR the way
        it was.}
        disable_interrupts;
        old_mcr := Port [uart_mcr];
        Port [uart_mcr] := $A OR (old_mcr AND 1);
        enable_interrupts;
        {Enable interrupt on data-available and interrupt upon change in
        modem status register (specifically, DSR).  The interrupt for
        transmit-ready is enabled when a character is put into the
        transmit queue, and disabled when the transmit queue is empty.}
        Port [uart_ier] := $09;
        {Enable the interrupt controller for this interrupt.}
        disable_interrupts;
        Port [$21] := Port [$21] AND NOT i8259bit;
        enable_interrupts;
        End;
      End;
  End;


{Deinstall the interrupt driver completely.  It doesn't change the baud rate
or mess with DTR; it tries to leave the interrupt vectors and enables and
everything else as it was when the driver was installed.

This Procedure MUST be called by the exit Procedure of this module before
the program exits to DOS, or the interrupt driver will still
be attached to its vector -- the next communications interrupt that came
along would jump to the interrupt driver which is no longer protected and
may have been written over.}


Procedure com_deinstall;
  Begin
  If com_installed Then
    Begin
    com_installed := False;
    Dispose(Rx_Queue);
    Dispose(Tx_Queue);
    {Restore Modem-Control-Register and Interrupt-Enable-Register.}
    Port [uart_mcr] := old_mcr;
    Port [uart_ier] := old_ier;
    {Restore appropriate bit of interrupt controller's mask}
    disable_interrupts;
    Port [$21] := Port [$21] AND NOT i8259bit OR
     old_i8259_mask AND i8259bit;
    enable_interrupts;
    {Reset the interrupt vector}
    SetIntVec (intnum, old_vector);
    End;
  End;



Constructor DirComObj.Init;
  Begin
  coBaud := 2400;
  coPort := 1;
  coFlow := 0;
  coParity := 0;
  coNumBits := 8;
  coStopBits := 1;
  End;


Destructor DirComObj.Done;
  Begin
  Com_DeInstall;
  End;


Function DirComObj.OpenPort(PortNo: Byte; Baud: LongInt; Flow: Byte; NumBits: Byte;
  StopBits: Byte; Parity: Byte): Boolean;
  Var
    Error: Word;

  Begin
  coPort := PortNo;
  coBaud := Baud;
  coFlow := Flow;
  coNumBits := NumBits;
  coStopBits := StopBits;
  coParity := Parity;
  com_install(coPort, Error);
  OpenPort := Error = 0;
  End;


Procedure DirComObj.ClosePort;
  Begin
  Com_DeInstall;
  End;


Function DirComObj.GetBPS: LongInt;
  Begin
  GetBPS := coBaud;
  End;


Function DirComObj.GetInfo: String;
  Begin
  GetInfo := 'IBMCOM Version 3.2 by Wayne Conrad';
  End;


Function DirComObj.SendCharNW(Ch: Char): Boolean;
  Begin
  If OutReady Then
    Begin
    SendChar(Ch);
    SendCharNW := True;
    End
  Else
    SendCharNW := False;
  End;


Procedure DirComObj.SetFlowCtl(FC: Byte);
  Begin
  End;

Procedure DirComObj.StartBreak;
  Begin
  End;

Procedure DirComObj.StopBreak;
  Begin
  End;


Function DirComObj.SendBlock(Var Dt; DSize: Word): Word;
  Var
    CharsAvail: Word;
    i: Word;

  Type CharArray = Array[1..$f000] of Byte;

  Begin
  If com_installed Then
    Begin
    disable_interrupts;
    CharsAvail := Tx_Queue_Size - Tx_Chars;
    If CharsAvail > DSize Then
      CharsAvail := DSize;
    SendBlock := CharsAvail;
    i := 1;
    While i <= CharsAvail Do
      Begin
      tx_queue^ [tx_in] := CharArray(Dt)[i];
      If tx_in < tx_queue_size Then
        Inc (tx_in)
      Else
        tx_in := 1;
      Inc (tx_chars);
      Inc(i);
      End;
    If Odd (Port [uart_msr] Shr 4) Then
      Port [uart_ier] := Port [uart_ier] Or 2;
    enable_interrupts;
    End
  Else
    SendBlock := 0;
  End;


Function DirComObj.GetBlock(Var Dt; DSize: Word): Word;
  Var
    CharsAvail: Word;
    i: Word;

  Type CharArray = Array[1..$f000] of Byte;

  Begin
  If Not com_installed Or (rx_chars = 0) Then
    GetBlock := 0
  Else
    Begin
    disable_interrupts;
    CharsAvail := Rx_Queue_Size - Rx_Chars;
    If CharsAvail > DSize Then
      CharsAvail := DSize;
    GetBlock := CharsAvail;
    i := 1;
    While i <= CharsAvail Do
      Begin
      CharArray(Dt)[i] := rx_queue^ [rx_out];
      Inc (rx_out);
      If rx_out > rx_queue_size Then
        rx_out := 1;
      Dec (rx_chars);
      Inc(i);
      End;
    enable_interrupts;
    End;
  End;



{This Procedure is called when the program exits for any reason.  It
deinstalls the interrupt driver.}
{$F+}
Procedure exit_Procedure;
  Begin
  com_deinstall;
  ExitProc := exit_save;
  End;



Begin {This installs the exit Procedure.}
Exit_Save := ExitProc;
ExitProc := @exit_Procedure;
End.
